"""Handle the paths of the analysis products."""

import logging
from pathlib import Path
from typing import List

from astropy.table import Table
from lstchain.onsite import find_systematics_correction_file, find_time_calibration_file
from lstchain.scripts.onsite.onsite_create_calibration_file import search_filter

from osa.configs import options
from osa.configs.config import DEFAULT_CFG
from osa.configs.config import cfg
from osa.configs.datamodel import Sequence
from osa.utils import utils
from osa.utils.logging import myLogger

log = myLogger(logging.getLogger(__name__))

__all__ = [
    "get_calibration_file",
    "get_drs4_pedestal_file",
    "pedestal_ids_file_exists",
    "get_run_date",
    "drs4_pedestal_exists",
    "calibration_file_exists",
    "sequence_calibration_files",
    "destination_dir",
    "datacheck_directory",
    "get_datacheck_files",
    "get_drive_file",
    "get_summary_file",
    "get_pedestal_ids_file",
    "DATACHECK_WEB_BASEDIR",
    "DEFAULT_CFG",
    "create_source_directories",
    "analysis_path",
]


DATACHECK_WEB_BASEDIR = Path(cfg.get("WEBSERVER", "DATACHECK"))


def analysis_path(tel) -> Path:
    """
    Path of the running_analysis directory for a certain date

    Returns
    -------
    directory : Path
        Path of the running_analysis directory for a certain date
    """
    log.debug(f"Getting analysis path for telescope {tel}")
    flat_date = utils.date_to_dir(options.date)
    options.prod_id = utils.get_prod_id()
    directory = Path(cfg.get(tel, "ANALYSIS_DIR")) / flat_date / options.prod_id

    if not options.simulate:
        directory.mkdir(parents=True, exist_ok=True)
    else:
        log.debug("SIMULATE the creation of the analysis directory.")

    log.debug(f"Analysis directory: {directory}")
    return directory


def get_run_date(run_id: int) -> str:
    """
    Return the date (YYYYMMDD) when the given run was taken. The search for this date
    is done by looking at the date corresponding to each run in the merged run summaries
    file.
    """
    merged_run_summaries_file = cfg.get("LST1", "MERGED_SUMMARY")
    summary_table = Table.read(merged_run_summaries_file)

    try:
        date_string = summary_table[summary_table["run_id"] == run_id]["date"][0]
    except IndexError:
        log.warning(
            f"Run {run_id} is not in the summary table. "
            f"Assuming the date of the run is {options.date}."
        )
        date_string = utils.date_to_dir(options.date)

    return date_string.replace("-", "")


def get_drs4_pedestal_file(run_id: int) -> Path:
    """
    Return the drs4 pedestal file corresponding to a given run id
    regardless of the date when the run was taken.
    """
    drs4_pedestal_dir = Path(cfg.get("LST1", "PEDESTAL_DIR"))
    date = get_run_date(run_id)
    file = drs4_pedestal_dir / date / f"pro/drs4_pedestal.Run{run_id:05d}.0000.h5"
    return file.resolve()


def get_calibration_file(run_id: int) -> Path:
    """
    Return the calibration file corresponding to a given run_id.

    Parameters
    ----------
    run_id : int
        Run id of the calibration file to be built.

    Notes
    -----
    The file path will be built regardless of the date when the run was taken.
    We follow the naming convention of the calibration files produced by the lstchain script
    which depends on the filter wheels position. Therefore, we need to try to fetch the filter
    position from the CaCo database. If the filter position is not found, we assume the default
    filter position 5-2. Filter information is not available in the database for runs taken before
    mid 2021 approx.
    """

    calib_dir = Path(cfg.get("LST1", "CALIB_DIR"))
    date = get_run_date(run_id)

    if options.test:  # Run tests avoiding the access to the database
        filters = 52

    else:
        mongodb = cfg.get("database", "CaCo_db")
        try:
            # Cast run_id to int to avoid problems with numpy int64 encoding in MongoDB
            filters = search_filter(int(run_id), mongodb)
        except IOError:
            log.warning("No filter information found in database. Assuming positions 52.")
            filters = 52

    file = calib_dir / date / f"pro/calibration_filters_{filters}.Run{run_id:05d}.0000.h5"

    return file.resolve()


def pedestal_ids_file_exists(run_id: int) -> bool:
    """Look for the files with pedestal interleaved event identification."""
    pedestal_ids_dir = Path(cfg.get("LST1", "PEDESTAL_FINDER_DIR"))
    file_list = sorted(pedestal_ids_dir.rglob(f"pedestal_ids_Run{run_id:05d}.*.h5"))
    return bool(file_list)


def drs4_pedestal_exists(run_id: int) -> bool:
    """Return true if drs4 pedestal file was already produced."""
    file = get_drs4_pedestal_file(run_id)
    return file.exists()


def calibration_file_exists(run_id: int) -> bool:
    """Return true if calibration file was already produced."""
    file = get_calibration_file(run_id)
    return file.exists()


def get_drive_file(date: str) -> Path:
    """Return the drive file corresponding to a given date in YYYYMMDD format."""
    yy_mm_dd = utils.date_in_yymmdd(date)
    drive_dir = Path(cfg.get("LST1", "DRIVE_DIR"))
    return (drive_dir / f"drive_log_{yy_mm_dd}.txt").resolve()


def get_summary_file(date) -> Path:
    """Return the run summary file corresponding to a given date in YYYYMMDD format."""
    run_summary_dir = Path(cfg.get("LST1", "RUN_SUMMARY_DIR"))
    return (run_summary_dir / f"RunSummary_{date}.ecsv").resolve()


def get_pedestal_ids_file(run_id: int, date: str) -> Path:
    """
    Return the pedestal ids file path corresponding to a given run
    from a date in format YYYYMMDD.
    """
    pedestal_ids_dir = Path(cfg.get("LST1", "PEDESTAL_FINDER_DIR")) / date
    file = pedestal_ids_dir / f"pedestal_ids_Run{run_id:05d}.{{subruns:04d}}.h5"
    return file.resolve()


def sequence_calibration_files(sequence_list: List[Sequence]) -> None:
    """Build names of the calibration files for each sequence in the list."""
    flat_date = utils.date_to_dir(options.date)
    base_dir = Path(cfg.get("LST1", "BASE"))

    for sequence in sequence_list:
        # Assign the calibration files to the sequence object
        sequence.drs4_file = get_drs4_pedestal_file(sequence.drs4_run)
        sequence.calibration_file = get_calibration_file(sequence.pedcal_run)
        sequence.time_calibration_file = find_time_calibration_file(
            "pro", sequence.pedcal_run, base_dir=base_dir
        )
        sequence.systematic_correction_file = find_systematics_correction_file(
            "pro", flat_date, base_dir=base_dir
        )


def get_datacheck_files(pattern: str, directory: Path) -> list:
    """Return a list of files matching the pattern."""
    return sorted(directory.glob(pattern))


def datacheck_directory(data_type: str, date: str) -> Path:
    """Returns the path to the datacheck directory given the data type."""
    if data_type in {"PEDESTAL", "CALIB"}:
        directory = Path(cfg.get("LST1", f"{data_type}_DIR")) / date / "pro/log"
    elif data_type == "DL1AB":
        directory = destination_dir("DL1AB", create_dir=False)
    elif data_type == "LONGTERM":
        directory = Path(cfg.get("LST1", f"{data_type}_DIR")) / options.prod_id / date
    else:
        raise ValueError(f"Unknown data type: {data_type}")
    return directory


def destination_dir(concept: str, create_dir: bool = True) -> Path:
    """
    Create final destination directory for each data level.
    See Also osa.utils.register_run_concept_files

    Parameters
    ----------
    concept : str
        Expected: MUON, DL1AB, DATACHECK, DL2, PEDESTAL, CALIB, TIMECALIB
    create_dir : bool
        Set it to True (default) if you want to create the directory.
        Otherwise, it just returns the path

    Returns
    -------
    path : pathlib.Path
        Path to the directory
    """
    nightdir = utils.date_to_dir(options.date)

    if concept == "MUON":
        directory = Path(cfg.get(options.tel_id, concept + "_DIR")) / nightdir / options.prod_id
    elif concept in {"DL1AB", "DATACHECK"}:
        directory = (
            Path(cfg.get(options.tel_id, concept + "_DIR"))
            / nightdir
            / options.prod_id
            / options.dl1_prod_id
        )
    elif concept in {"DL2", "DL3"}:
        directory = (
            Path(cfg.get(options.tel_id, concept + "_DIR"))
            / nightdir
            / options.prod_id
            / options.dl2_prod_id
        )
    elif concept in {"PEDESTAL", "CALIB", "TIMECALIB"}:
        directory = (
            Path(cfg.get(options.tel_id, concept + "_DIR")) / nightdir / options.calib_prod_id
        )
    elif concept == "HIGH_LEVEL":
        directory = Path(cfg.get(options.tel_id, concept + "_DIR")) / nightdir / options.prod_id
    else:
        log.warning(f"Concept {concept} not known")
        directory = None

    if not options.simulate and create_dir:
        log.debug(f"Destination directory created for {concept}: {directory}")
        directory.mkdir(parents=True, exist_ok=True)
    else:
        log.debug(f"SIMULATING creation of final directory for {concept}")

    return directory


def create_source_directories(source_list: list, cuts_dir: Path):
    """Create a subdirectory for each source."""
    for source in source_list:
        if source is not None:
            source_dir = cuts_dir / source
            source_dir.mkdir(parents=True, exist_ok=True)



"""Utility functions for OSA pipeline provenance."""

import logging
import os
import re
import sys
from pathlib import Path

from osa.configs import options
from osa.configs.config import cfg
from osa.utils.logging import myLogger
from osa.utils.utils import get_lstchain_version, date_to_dir

__all__ = ["parse_variables", "get_log_config"]

REDUCTION_TASKS = ["r0_to_dl1", "dl1ab", "dl1_datacheck", "dl1_to_dl2"]


def parse_variables(class_instance):
    """Parse variables needed in model"""

    # calibration_pipeline.py
    # -c sequencer.cfg
    # -d 2020_02_18
    # --drs4-pedestal-run 01804
    # --pedcal-run 01805
    # LST1

    # datasequence.py
    # -c sequencer.cfg
    # -d 2020_02_18
    # --prod-id v0.4.3_v00
    # --pedcal-file .../20200218/pro/calibration_filters_52.Run02006.0000.h5
    # --drs4-pedestal-file .../20200218/pro/drs4_pedestal.Run02005.0000.h5
    # --time-calib-file .../20191124/pro/time_calibration.Run01625.0000.h5
    # --drive-file .../lapp/DrivePositioning/drive_log_20_02_18.txt
    # --run-summary .../monitoring/RunSummary/RunSummary_20200101.ecsv
    # 02006.0000
    # LST1

    flat_date = date_to_dir(options.date)
    configfile_dl1b = cfg.get("lstchain", "dl1b_config")
    configfile_dl2 = cfg.get("lstchain", "dl2_config")
    raw_dir = cfg.get("LST1", "R0_DIR")
    rf_models_directory = cfg.get("lstchain", "RF_MODELS")
    dl1_dir = cfg.get("LST1", "DL1_DIR")
    dl2_dir = cfg.get("LST1", "DL2_DIR")
    calib_dir = cfg.get("LST1", "CALIB_DIR")
    pedestal_dir = cfg.get("LST1", "PEDESTAL_DIR")
    # summary_dir = cfg.get("LST1", "RUN_SUMMARY_DIR")
    # calib_base_dir = cfg.get("LST1", "CALIB_BASE_DIR")
    # sys_dir = calib_base_dir / "ffactor_systematics"
    class_instance.SoftwareVersion = get_lstchain_version()
    class_instance.ProcessingConfigFile = options.configfile
    class_instance.ObservationDate = flat_date
    if class_instance.__name__ in REDUCTION_TASKS:
        muon_dir = Path(dl1_dir) / flat_date / options.prod_id
        outdir_dl1 = Path(dl1_dir) / flat_date / options.prod_id / options.dl1_prod_id
        outdir_dl2 = Path(dl2_dir) / flat_date / options.prod_id / options.dl2_prod_id

    if class_instance.__name__ in ["drs4_pedestal", "calibrate_charge"]:
        # drs4_pedestal_run_id  [0] 1804
        # pedcal_run_id         [1] 1805
        # history_file          [2] .../20210913/v0.7.5/sequence_LST1_01805.0000.history
        class_instance.PedestalRun = f"{class_instance.args[0]:05d}"
        class_instance.CalibrationRun = f"{class_instance.args[1]:05d}"

        pro = "pro"
        # TODO - massive reprocessing vs. next day processing

        # according to code in onsite scripts in lstchain
        class_instance.RawObservationFilePedestal = os.path.realpath(
            f"{raw_dir}/{flat_date}/LST-1.1.Run{class_instance.args[0]:05d}.fits.fz"
        )
        class_instance.RawObservationFileCalibration = os.path.realpath(
            f"{raw_dir}/{flat_date}/LST-1.1.Run{class_instance.args[1]:05d}.fits.fz"
        )
        class_instance.PedestalCheckPlot = os.path.realpath(
            f"{pedestal_dir}/{flat_date}/{pro}/"
            f"log/drs4_pedestal.Run{class_instance.args[0]:05d}.0000.pdf"
        )
        class_instance.CalibrationCheckPlot = os.path.realpath(
            f"{calib_dir}/{flat_date}/{pro}/"
            f"log/calibration_filters_52.Run{class_instance.args[1]:05d}.0000.pdf"
        )

        # according to code in sequence_calibration_files from paths.py
        class_instance.PedestalFile = os.path.realpath(
            f"{pedestal_dir}/{flat_date}/{pro}/"
            f"drs4_pedestal.Run{class_instance.args[0]:05d}.0000.h5"
        )
        class_instance.CoefficientsCalibrationFile = os.path.realpath(
            f"{calib_dir}/{flat_date}/{pro}/"
            f"calibration_filters_52.Run{class_instance.args[1]:05d}.0000.h5"
        )

    if class_instance.__name__ == "r0_to_dl1":
        # calibration_file   [0] .../20200218/pro/calibration_filters_52.Run02006.0000.h5
        # drs4_pedestal_file [1] .../20200218/pro/drs4_pedestal.Run02005.0000.h5
        # time_calib_file    [2] .../20191124/pro/time_calibration.Run01625.0000.h5
        # systematic_corr    [3] .../20200101/pro/no_sys_corrected_calib_20210514.0000.h5
        # drive_file         [4] .../DrivePositioning/drive_log_20_02_18.txt
        # run_summary_file   [5] .../RunSummary/RunSummary_20200101.ecsv
        # pedestal_ids_file  [6] .../path/to/interleaved/pedestal/events.h5
        # run_str            [7] 02006.0000

        run_subrun_id = class_instance.args[7]
        class_instance.ObservationRun = run_subrun_id.split(".")[0]
        # use realpath to resolve symbolic links and return abspath
        calibration_file = os.path.realpath(class_instance.args[0])
        pedestal_file = os.path.realpath(class_instance.args[1])
        timecalibration_file = os.path.realpath(class_instance.args[2])
        systematic_correction_file = os.path.realpath(class_instance.args[3])
        class_instance.R0SubrunDataset = os.path.realpath(
            f"{raw_dir}/{flat_date}/LST-1.1.Run{run_subrun_id}.fits.fz"
        )
        class_instance.CoefficientsCalibrationFile = calibration_file
        class_instance.PedestalFile = pedestal_file
        class_instance.TimeCalibrationFile = timecalibration_file
        class_instance.SystematicCorrectionFile = systematic_correction_file
        class_instance.PointingFile = os.path.realpath(class_instance.args[4])
        class_instance.RunSummaryFile = os.path.realpath(class_instance.args[5])
        class_instance.DL1SubrunDataset = os.path.realpath(
            f"{outdir_dl1}/dl1_LST-1.Run{run_subrun_id}.h5"
        )
        class_instance.MuonsSubrunDataset = os.path.realpath(
            f"{muon_dir}/muons_LST-1.Run{run_subrun_id}.fits"
        )
        class_instance.InterleavedPedestalEventsFile = None
        if class_instance.args[6] is not None:
            class_instance.InterleavedPedestalEventsFile = os.path.realpath(class_instance.args[6])

    if class_instance.__name__ == "dl1ab":
        # run_str       [0] 02006.0000

        class_instance.Analysisconfigfile_dl1 = os.path.realpath(configfile_dl1b)
        class_instance.ObservationRun = class_instance.args[0].split(".")[0]
        class_instance.StoreImage = cfg.getboolean("lstchain", "store_image_dl1ab")
        class_instance.DL1SubrunDataset = os.path.realpath(
            f"{outdir_dl1}/dl1_LST-1.Run{class_instance.args[0]}.h5"
        )

    if class_instance.__name__ == "dl1_datacheck":
        # run_str       [0] 02006.0000

        class_instance.ObservationRun = class_instance.args[0].split(".")[0]
        class_instance.DL1SubrunDataset = os.path.realpath(
            f"{outdir_dl1}/dl1_LST-1.Run{class_instance.args[0]}.h5"
        )
        class_instance.MuonsSubrunDataset = os.path.realpath(
            f"{muon_dir}/muons_LST-1.Run{class_instance.args[0]}.fits"
        )
        class_instance.DL1CheckSubrunDataset = os.path.realpath(
            f"{outdir_dl1}/datacheck_dl1_LST-1.Run{class_instance.args[0]}.h5"
        )
        class_instance.DL1CheckHDF5File = os.path.realpath(
            f"{outdir_dl1}/datacheck_dl1_LST-1.Run{class_instance.ObservationRun}.h5"
        )
        class_instance.DL1CheckPDFFile = os.path.realpath(
            f"{outdir_dl1}/datacheck_dl1_LST-1.Run{class_instance.ObservationRun}.pdf"
        )

    if class_instance.__name__ == "dl1_to_dl2":
        # run_str       [0] 02006.0000

        class_instance.Analysisconfigfile_dl2 = configfile_dl2
        class_instance.ObservationRun = class_instance.args[0].split(".")[0]
        class_instance.RFModelEnergyFile = os.path.realpath(f"{rf_models_directory}/reg_energy.sav")
        class_instance.RFModelDispNormFile = os.path.realpath(
            f"{rf_models_directory}/reg_disp_norm.sav"
        )
        class_instance.RFModelDispSignFile = os.path.realpath(
            f"{rf_models_directory}/reg_disp_sign.sav"
        )
        class_instance.RFModelGammanessFile = os.path.realpath(f"{rf_models_directory}/cls_gh.sav")
        class_instance.DL1SubrunDataset = os.path.realpath(
            f"{outdir_dl1}/dl1_LST-1.Run{class_instance.args[0]}.h5"
        )
        class_instance.DL2SubrunDataset = os.path.realpath(
            f"{outdir_dl2}/dl2_LST-1.Run{class_instance.args[0]}.h5"
        )
        class_instance.DL2MergedFile = os.path.realpath(
            f"{outdir_dl2}/dl2_LST-1.Run{class_instance.ObservationRun}.h5"
        )

    if class_instance.__name__ in REDUCTION_TASKS:
        class_instance.session_name = class_instance.ObservationRun

    return class_instance


def get_log_config():
    """Get logging configuration from an OSA config file."""
    # Default config filename value
    config_file = Path(__file__).resolve().parent / ".." / ".." / options.configfile
    std_logger_file = Path(__file__).resolve().parent / "config" / "logger.yaml"

    # fetch config filename value from args
    in_config_arg = False
    for arg in sys.argv:
        if in_config_arg:
            config_file = arg
            in_config_arg = False
        if arg in ["-c", "--config"]:
            in_config_arg = True

    # parse configuration
    log_config = ""
    in_prov_section = False
    str_path_tests = str(Path(__file__).resolve().parent / "tests" / "prov.log")
    try:
        with open(config_file, "r") as f:
            for line in f.readlines():
                if "pytest" in sys.modules and in_prov_section:
                    line = re.sub(r"filename:(.*)$", f"filename: {str_path_tests}", line)
                if in_prov_section:
                    log_config += line
                if "[PROVENANCE]" in line:
                    in_prov_section = True
    except FileNotFoundError:
        log = myLogger(logging.getLogger(__name__))
        log.warning(f"{config_file} not found, using {std_logger_file} instead.")

    # use default logger.yaml if no prov config info found
    if log_config == "":
        log_config = std_logger_file.read_text()

    return log_config



"""Query the TCU database source name and astronomical coordinates."""
import logging
from datetime import datetime

from pymongo import MongoClient
from pymongo.errors import ConnectionFailure

from osa.utils.logging import myLogger

__all__ = ["query", "db_available"]


log = myLogger(logging.getLogger(__name__))


def db_available():
    """Check the connection to the TCU database."""
    caco_client = MongoClient("tcs01", serverSelectionTimeoutMS=3000)
    tcu_client = MongoClient("tcs05", serverSelectionTimeoutMS=3000)
    try:
        caco_client.server_info()
        tcu_client.server_info()
    except ConnectionFailure:
        log.info("TCU or CaCo database not available. No source info will be added.")
        return False
    else:
        return True


def query(obs_id: int, property_name: str):
    """
    Query the source name and coordinates from TCU database.

    Parameters
    ----------
    obs_id : int
        Run number
    property_name : str
        Properties from drive information e.g. `DriveControl_SourceName`,
        `DriveControl_RA_Target`, `DriveControl_Dec_Target`

    Returns
    -------
    query_result : str or None
        Query result from database. It can be either the source name or its coordinates.

    Raises
    ------
    ConnectionFailure
    """

    # Avoid problems with numpy int64 encoding in MongoDB
    if not isinstance(obs_id, int):
        obs_id = int(obs_id)

    caco_client = MongoClient("tcs01")
    tcu_client = MongoClient("tcs05")

    with caco_client, tcu_client:
        run_info = caco_client["CACO"]["RUN_INFORMATION"]
        run = run_info.find_one({"run_number": obs_id})

        try:
            start = datetime.fromisoformat(run["start_time"].replace("Z", ""))
            end = datetime.fromisoformat(run["stop_time"].replace("Z", ""))
        except TypeError:
            return None

        bridges_monitoring = tcu_client["bridgesmonitoring"]
        property_collection = bridges_monitoring["properties"]
        chunk_collection = bridges_monitoring["chunks"]
        descriptors = property_collection.find(
            {"property_name": property_name},
        )

        entries = {"name": property_name, "time": [], "value": []}

        for descriptor in descriptors:
            query_property = {"pid": descriptor["_id"]}

            if start is not None:
                query_property["begin"] = {"$gte": start}

            if end is not None:
                query_property["end"] = {"$lte": end}

            chunks = chunk_collection.find(query_property)

            for chunk in chunks:
                for value in chunk["values"]:
                    entries["time"].append(value["t"])
                    entries["value"].append(value["val"])

                    source_name = entries["value"][0]
                    return source_name if source_name != "" else None



from datetime import datetime


obs_date = datetime.fromisoformat("2020-01-17")


def test_source_list(
    r0_data,
    run_summary_file,
    run_catalog,
    drs4_time_calibration_files,
    systematic_correction_files,
    pedestal_ids_file,
    merged_run_summary,
):
    """Test that the list of name of sources is correct."""
    from osa.nightsummary.extract import get_source_list

    sources = get_source_list(date=obs_date)
    assert list(sources) == ["Crab", "MadeUpSource"]
    assert sources == {"Crab": [1807], "MadeUpSource": [1808]}


def test_build_sequences(sequence_list):
    """Test that building of sequences."""
    from osa.nightsummary.extract import build_sequences

    extracted_seq_list = build_sequences(date=obs_date)
    for sequence, extracted_seq in zip(sequence_list, extracted_seq_list):
        assert sequence.run == extracted_seq.run
